// Copyright
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Underlying paper:
//
// Layer-Parallel Training of Deep Residual Neural Networks
// S. Guenther, L. Ruthotto, J.B. Schroder, E.C. Czr, and N.R. Gauger
//
// Download: https://arxiv.org/pdf/1812.04352.pdf
//

#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/resource.h>

#include "braid_wrapper.hpp"
#include "config.hpp"
#include "dataset.hpp"
#include "defs.hpp"
#include "hessianApprox.hpp"
#include "layer.hpp"
#include "network.hpp"
#include "util.hpp"

#define MASTER_NODE 0

int main(int argc, char *argv[]) {
  /* --- Data --- */
  Config *config;          /**< Storing configurations */
  DataSet *trainingdata;   /**< Training dataset */
  DataSet *validationdata; /**< Validation dataset */

  /* --- Network --- */
  Network *network; /**< DNN Network architecture */
  int startlayerID, endlayerID; /**< Index of first and last layer stored on this processor */
  MyReal accur_train = 0.0; /**< Accuracy on training data */
  MyReal accur_val = 0.0;   /**< Accuracy on validation data */
  MyReal loss_train = 0.0;  /**< Loss function on training data */
  MyReal loss_val = 0.0;    /**< Loss function on validation data */
  MyReal losstrain_out = 0.0;
  MyReal lossval_out = 0.0;
  MyReal accurtrain_out = 0.0;
  MyReal accurval_out = 0.0;

  /* --- XBraid --- */
  myBraidApp *primaltrainapp;         /**< Braid App for training data */
  myAdjointBraidApp *adjointtrainapp; /**< Adjoint Braid for training data */
  myBraidApp *primalvalapp;           /**< Braid App for validation data */

  /* --- Optimization --- */
  int ndesign_local;  /**< Number of local design variables on this processor */
  int ndesign_global; /**< Number of global design variables (sum of local)*/
  MyReal *ascentdir = 0; /**< Direction for design updates */
  MyReal objective;      /**< Optimization objective */
  MyReal wolfe;          /**< Holding the wolfe condition value */
  MyReal rnorm;          /**< Space-time Norm of the state variables */
  MyReal rnorm_adj;      /**< Space-time norm of the adjoint variables */
  MyReal gnorm;          /**< Norm of the gradient */
  MyReal ls_param;       /**< Parameter in wolfe condition test */
  MyReal stepsize;       /**< Stepsize used for design update */
  char optimfilename[255];
  FILE *optimfile = 0;
  char loss_filename[255];
  FILE *loss_file = 0;
  MyReal ls_stepsize;
  MyReal ls_objective, test_obj;
  int ls_iter;

  /* --- Time measurements --- */
  struct rusage r_usage;
  MyReal StartTime, StopTime, myMB, globalMB;
  MyReal UsedTime = 0.0;

  MyReal loss_MAPE_train = 0.0;
  MyReal loss_MAPE_val = 0.0; 
  MyReal prev_vl = 0.0;
  MyReal loss_MAPE_val_out = 0.0;
  MyReal loss_MAPE_train_out = 0.0;
  int patience, counter;


  /* Initialize MPI */
  int myid;
  int size;
  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &myid);
  MPI_Comm_size(MPI_COMM_WORLD, &size);

  /* Instantiate objects */
  config = new Config();
  trainingdata = new DataSet();
  validationdata = new DataSet();
  network = new Network(MPI_COMM_WORLD);

  /* Read config file */
  if (argc != 2) {
    if (myid == MASTER_NODE) {
      printf("\n");
      printf("USAGE: ./main </path/to/configfile> \n");
    }
    MPI_Finalize();
    return 0;
  }

  int err = config->readFromFile(argv[1]);
  if (err) {
    printf("Error while reading config file!\n");
    MPI_Finalize();
    return 0;
  }

  /* Caution: Undo the following changes later!!! 
  Making temporary changes here to enable debugging! */
  /* 
  if (argc != 1) {
    if (myid == MASTER_NODE) {
      printf("\n");
      printf("USAGE: ./main </path/to/configfile> \n");
    }
    MPI_Finalize();
    return 0;
  }
  char** path_file = examples/SSHE_Analytical/8/SSHE.cfg
  int err = config->readFromFile(*path_file);
  if (err) {
    printf("Error while reading config file!\n");
    MPI_Finalize();
    return 0;
  }
  */


  /* Initialize training and validation data */
  trainingdata->initialize(config->ntraining, config->nfeatures,
                           config->nclasses, config->nbatch, MPI_COMM_WORLD);
  trainingdata->readData(config->datafolder, config->ftrain_ex,
                         config->ftrain_labels);

  validationdata->initialize(config->nvalidation, config->nfeatures,
                             config->nclasses, config->nvalidation,
                             MPI_COMM_WORLD);  // full validation set!
  validationdata->readData(config->datafolder, config->fval_ex,
                           config->fval_labels);

  /* Initialize XBraid */
  primaltrainapp =
      new myBraidApp(trainingdata, network, config, MPI_COMM_WORLD);
  adjointtrainapp = new myAdjointBraidApp(
      trainingdata, network, config, primaltrainapp->getCore(), MPI_COMM_WORLD);
  primalvalapp =
      new myBraidApp(validationdata, network, config, MPI_COMM_WORLD);
  primaltrainapp->GetGridDistribution(&startlayerID, &endlayerID);
  if (startlayerID == 0) startlayerID = startlayerID - 1; // -1 is index of the opening layer

  /* Initialize the network  */
  network->createLayerBlock(startlayerID, endlayerID, config);
  network->setDesignRandom(config->weights_open_init, config->weights_init, config->weights_class_init);
  network->setDesignFromFile(config->datafolder, config->weightsopenfile, NULL, config->weightsclassificationfile);
  ndesign_local = network->getnDesignLocal();
  ndesign_global = network->getnDesignGlobal();

  /* Print some neural network information */
  printf("%d: Layer range: [%d, %d] / %d\n", myid, startlayerID, endlayerID,
         config->nlayers);
  printf("%d: Design variables (local/global): %d/%d\n", myid, ndesign_local,
         ndesign_global);

  /* Initialize Hessian approximation */
  HessianApprox *hessian = NULL;
  switch (config->hessianapprox_type) {
    case BFGS_SERIAL:
      hessian = new BFGS(MPI_COMM_WORLD, ndesign_local);
      break;
    case LBFGS:
      hessian = new L_BFGS(MPI_COMM_WORLD, ndesign_local, config->lbfgs_stages);
      break;
    case IDENTITY:
      hessian = new Identity(MPI_COMM_WORLD, ndesign_local);
      break;
    default:
      printf("Error: unexpected hessianapprox_type returned");
      return 0;
  }

  /* Initialize optimization parameters */
  ascentdir = new MyReal[ndesign_local];
  stepsize = config->getStepsize(0);
  gnorm = 0.0;
  objective = 0.0;
  rnorm = 0.0;
  rnorm_adj = 0.0;
  ls_param = 1e-4;
  ls_iter = 0;
  ls_stepsize = stepsize;

  /* Open and prepare optimization and loss output file*/
  if (myid == MASTER_NODE) {
    sprintf(optimfilename, "%s.dat", "optim");
    optimfile = fopen(optimfilename, "w");
    config->writeToFile(optimfile);
    sprintf(loss_filename, "%s.dat", "loss_history");    
    loss_file = fopen(loss_filename, "w");
    fprintf(optimfile,
            "#    || r ||          || r_adj ||      Objective             Loss "
            "                 || grad ||            Stepsize  ls_iter   "
            "Accur_train  Accur_val   Time(sec)\n");
  }

  /* Measure wall time */
  StartTime = MPI_Wtime();
  StopTime = 0.0;
  UsedTime = 0.0;
  
  /* Early Stopping Patience */
  patience = 20;
  counter = 0;
   
  /** Main optimization iteration
   *
   * The following loop represents the paper's Algorithm (2)
   *
   */
  for (int iter = 0; iter < config->maxoptimiter; iter++) {
    /* Set up the current batch */
    trainingdata->selectBatch(config->batch_type, MPI_COMM_WORLD);

    /** Solve state and adjoint equations (2.15) and (2.17)
     *
     *  Algorithm (2): Step 1 and 2
     */
    network->training_phase = 1;
    rnorm = primaltrainapp->run();
    rnorm_adj = adjointtrainapp->run();

    /* Get output */
    objective = primaltrainapp->getObjective();
    loss_train = network->getLoss();
    loss_MAPE_train = network->get_PAME();
    // accur_train = network->getAccuracy();

    /* --- Validation data: Get accuracy --- */
    if (config->validationlevel > 0) {
      network->training_phase = 0;/* Validation Phase */
      primalvalapp->run();
      loss_val = network->getLoss();
      loss_MAPE_val = network->get_PAME();
      /* Del: Not needed for regression
      accur_val = network->getAccuracy(); */
    }
    /* --- Optimization control and output ---*/

    /** Compute global gradient norm
     *
     *  Algorithm (2): Step 3
     */
    gnorm = vecnorm_par(ndesign_local, network->getGradient(), MPI_COMM_WORLD);

    /* Communicate loss and accuracy. This is actually only needed for output.
     * TODO: Remove it. */
    /* MPI ALl_Reduce will basically add all the loss_train from all */
    MPI_Allreduce(&loss_train, &losstrain_out, 1, MPI_MyReal, MPI_SUM,
                  MPI_COMM_WORLD);
    MPI_Allreduce(&loss_val, &lossval_out, 1, MPI_MyReal, MPI_SUM,
                  MPI_COMM_WORLD);
    MPI_Allreduce(&accur_train, &accurtrain_out, 1, MPI_MyReal, MPI_SUM,
                  MPI_COMM_WORLD);
    MPI_Allreduce(&accur_val, &accurval_out, 1, MPI_MyReal, MPI_SUM,
                  MPI_COMM_WORLD);
    MPI_Allreduce(&loss_MAPE_train, &loss_MAPE_train_out, 1, MPI_MyReal, MPI_SUM,
                  MPI_COMM_WORLD);
    MPI_Allreduce(&loss_MAPE_val, &loss_MAPE_val_out, 1, MPI_MyReal, MPI_SUM,
                  MPI_COMM_WORLD);
    
    /* Output */
    StopTime = MPI_Wtime();
    UsedTime = StopTime - StartTime;
    if (myid == MASTER_NODE) {
      
      printf(
          "\niter \t|| r ||\t\t|| r_adj ||\tObjective\t\tTraining_Loss\t\t\t|| grad "
          "||\t\tStepsize\tls_iter\tVal_Loss\tTime(sec)\n");
      printf(
          "%03d\t%1.8e\t%1.8e\t%1.14e\t%1.14e\t%1.14e\t%5f\t%2d\t%2.2f%%\t\t%2."
          "2f%%\t\t%.1f\n\n",
          iter, rnorm, rnorm_adj, objective, losstrain_out, gnorm, stepsize,
          ls_iter, lossval_out, UsedTime);
      /*
      fprintf(optimfile,
              "%03d  %1.8e  %1.8e  %1.14e  %1.14e  %1.14e  %5f  %2d        "
              "%2.2f%%      %2.2f%%     %.1f\n",
              iter, rnorm, rnorm_adj, objective, losstrain_out, gnorm, stepsize,
               ls_iter, accurtrain_out, accurval_out, UsedTime);
      fflush(optimfile);
      */
     
     /* New: Write the Training Loss + Validation Loss Vs Number of iterations */
            if (config->validationlevel > 0) {
        fprintf(loss_file,
              "%03d\t%1.14e\t%1.14e\t%1.14e\t%1.14e\n",
                // iter, loss_train, losstrain_out, loss_val,lossval_out);
              iter, losstrain_out, lossval_out,loss_MAPE_train_out, loss_MAPE_val_out );
        fflush(loss_file);
      }
      else
      {
        fprintf(loss_file,
              "%03d\t%1.14e\t%1.14e\n",
              iter, losstrain_out, loss_MAPE_train);
        fflush(loss_file);
      }
      


    }

    /** Check optimization convergence
     *
     *  Algorithm (2): Step 6
     */
    if (gnorm < config->gtol) {
      if (myid == MASTER_NODE) {
        printf("Optimization has converged. \n");
        printf("Be happy and go home!       \n");
      }
      break;
    }
    if (iter == config->maxoptimiter - 1) {
      if (myid == MASTER_NODE) {
        printf("\nMax. optimization iterations reached.\n");
      }
      break;
    }

    /* Early stopping based on validation loss   */
    if (iter > 0)
    {
        if (lossval_out >= prev_vl) {
            counter = counter + 1;
        }
        else
        {
            counter = 0;
        }
        if (counter >= patience)
        {
          if (myid == MASTER_NODE) {
            printf("Optimization has converged. \n");
            printf("Early Stopping: Validation loss has stopped decreasing! \n");
          }   
          break;
        }
    }
    /* Set validation loss from previous iteration */
    prev_vl = lossval_out;
    
    /* If optimization didn't converge, continue */

    /* --- Design update --- */

    /** Compute search direction
     *
     *  Algorithm (2): Step 4
     */
    hessian->updateMemory(iter, network->getDesign(), network->getGradient());
    hessian->computeAscentDir(iter, network->getGradient(), ascentdir);
    stepsize = config->getStepsize(iter);

    /** Update the design/network control parameter in negative ascent direction
     *  and perform backtracking linesearch.
     *
     *  Algorithm (2): Step 5
     */
    vec_axpy(ndesign_local, -1.0*stepsize, ascentdir, network->getDesign());
    network->MPI_CommunicateNeighbours();

    if (config->stepsize_type == BACKTRACKINGLS) {
      /* Compute wolfe condition */
      wolfe = vecdot_par(ndesign_local, network->getGradient(), ascentdir,
                         MPI_COMM_WORLD);

      /* Start linesearch iterations */
      ls_stepsize = config->getStepsize(iter);
      stepsize = ls_stepsize;
      for (ls_iter = 0; ls_iter < config->ls_maxiter; ls_iter++) {
        primaltrainapp->getCore()->SetPrintLevel(0);
        primaltrainapp->run();
        ls_objective = primaltrainapp->getObjective();
        primaltrainapp->getCore()->SetPrintLevel(config->braid_printlevel);

        test_obj = objective - ls_param * ls_stepsize * wolfe;
        if (myid == MASTER_NODE)
          printf("ls_iter = %d:\tls_objective = %1.14e\ttest_obj = %1.14e\n",
                 ls_iter, ls_objective, test_obj);
        /* Test the wolfe condition */
        if (ls_objective <= test_obj) {
          /* Success, use this new design */
          break;
        } else {
          /* Test for line-search failure */
          if (ls_iter == config->ls_maxiter - 1) {
            if (myid == MASTER_NODE)
              printf("\n\n   WARNING: LINESEARCH FAILED! \n\n");
            break;
          }

          /* Go back part of the step */
          vec_axpy(ndesign_local, (1.0 - config->ls_factor) * stepsize, ascentdir, network->getDesign());
          network->MPI_CommunicateNeighbours();

          /* Decrease the stepsize */
          ls_stepsize = ls_stepsize * config->ls_factor;
          stepsize = ls_stepsize;
        }
      }
    }
  }// Training ends here!!!!!!! 
  
  /* Testing/validation starts here */ 
  /* --- Run final validation and write prediction file --- */
  if (config->validationlevel > -1) {
    if (myid == MASTER_NODE) printf("\n --- Run final validation ---\n");
    //network.testing_phase = 1;
    network->final_validation_phase =1;
    primalvalapp->getCore()->SetPrintLevel(0);
    primalvalapp->run();
    loss_val = network->getLoss();

    printf("Final validation loss:  %2.2f%%\n", loss_val);
  }
  /* To enable saving weights I suppose! */
  // write_vector("design.dat", design, ndesign);

  /* Print some statistics */
  StopTime = MPI_Wtime();
  UsedTime = StopTime - StartTime;
  getrusage(RUSAGE_SELF, &r_usage);
  myMB = (MyReal)r_usage.ru_maxrss / 1024.0;
  MPI_Allreduce(&myMB, &globalMB, 1, MPI_MyReal, MPI_SUM, MPI_COMM_WORLD);

  // printf("%d; Memory Usage: %.2f MB\n",myid, myMB);
  if (myid == MASTER_NODE) {
    printf("\n");
    printf(" Used Time:        %.2f seconds\n", UsedTime);
    printf(" Global Memory:    %.2f MB\n", globalMB);
    printf(" Processors used:  %d\n", size);
    printf("\n");
  }

  /* Clean up XBraid */
  delete network;

  delete primaltrainapp;
  delete adjointtrainapp;
  delete primalvalapp;

  /* Delete optimization vars */
  delete hessian;
  delete[] ascentdir;

  /* Delete training and validation examples  */
  delete trainingdata;
  delete validationdata;

  /* Close optim and loss file */
  if (myid == MASTER_NODE) {
    fclose(optimfile);
    printf("Optimfile: %s\n", optimfilename);
    fclose(loss_file);
    printf("loss_file: %s\n",loss_filename);
  }

  delete config;

  MPI_Finalize();
  return 0;
}
